@@ -5,14 +5,185 @@
 
 #pragma once
 
+#include "flatpakcommon.h"
 #include "flatpakreference.h"
 
+#include <QAbstractItemModel>
 #include <QAbstractListModel>
 #include <QPointer>
 #include <QString>
 
+#include <optional>
+#include <variant>
+
 class FlatpakReference;
 
+/** For exporting enum to QML */
+class FlatpakPermissionsSectionType : public QObject
+{
+    Q_OBJECT
+
+public:
+    enum Type {
+        Basic,
+        Filesystems,
+        Advanced,
+        SubsystemsShared,
+        Sockets,
+        Devices,
+        Features,
+        SessionBus,
+        SystemBus,
+        Environment,
+    };
+    Q_ENUM(Type)
+};
+
+/**
+ * Entry of the `filesystems=` list.
+ *
+ * "Deny" access mode represents an entry prefixes with a `!` bang character.
+ * Other modes correspond to `:ro`, `:rw` and `:create` suffixes.
+ *
+ * Path should be without trailing `/` slash character.
+ * See flatpak-metadata(5) for possible values.
+ */
+class FlatpakFilesystemsEntry
+{
+public:
+    enum class AccessMode {
+        /** Make the given directory available read-only. */
+        ReadOnly,
+        /** Make the given directory available read/write. This is the default. */
+        ReadWrite,
+        /** Make the given directory available read/write, and create it if it does not already exist. */
+        Create,
+        /** Don't expose filesystem to app. */
+        Deny,
+    };
+
+    enum class PathMode {
+        Required,
+        Optional,
+        NoPath,
+    };
+
+    enum class FilesystemPrefix {
+        /** Path is required. */
+        Absolute,
+        /** Path is optional. */
+        Home,
+        /** Path is invalid. */
+        Host,
+        HostOs,
+        HostEtc,
+        /** Path is optional. */
+        XdgDesktop,
+        XdgDocuments,
+        XdgDownload,
+        XdgMusic,
+        XdgPictures,
+        XdgPublicShare,
+        XdgVideos,
+        XdgTemplates,
+        /** Path is optional. */
+        XdgCache,
+        XdgConfig,
+        XdgData,
+        /** Path is required. */
+        XdgRun,
+        // Future-proof fallback variant. Prefix is empty, so path mode is
+        // required, which also ensures that is will be non-empty.
+        Unknown,
+    };
+
+    // For Required path mode, fixed string is empty because it is meaningless.
+    // For Invalid path mode, second, prefix string is empty.
+    struct TableEntry {
+        FilesystemPrefix prefix;
+        PathMode mode;
+        // Fixed magic filesystem name
+        QLatin1String fixedString;
+        // Same as fixed, but with '/' appended.
+        QLatin1String prefixString;
+    };
+
+    // Default constructor is required for meta-type registration.
+    /** Default constructor. Creates an invalid entry. */
+    FlatpakFilesystemsEntry() = default;
+
+    /**
+     * Construct new entry directly from raw unvalidated data.
+     */
+    explicit FlatpakFilesystemsEntry(FilesystemPrefix prefix, AccessMode mode, const QString &path = QString());
+
+    /**
+     * Parse list entry into data structure. Might fail for various reasons,
+     * such as ill-formed access mode, so returns an optional value.
+     */
+    static std::optional<FlatpakFilesystemsEntry> parse(QStringView entry);
+
+    /**
+     * Format this entry into string. Opposite of parse(). Omits default `:rw` suffix.
+     */
+    QString format() const;
+
+    FilesystemPrefix prefix() const;
+    QString path() const;
+    AccessMode mode() const;
+
+    // TODO C++20: use `= default` implementation.
+    bool operator==(const FlatpakFilesystemsEntry &other) const;
+    bool operator!=(const FlatpakFilesystemsEntry &other) const;
+
+private:
+    FilesystemPrefix m_prefix = FilesystemPrefix::Absolute;
+    AccessMode m_mode = AccessMode::ReadWrite;
+    // Depending on prefix type, path can be optional, required or illegal.
+    QString m_path;
+};
+
+Q_DECLARE_METATYPE(FlatpakFilesystemsEntry)
+
+class PolicyChoicesModel : public QAbstractListModel
+{
+    Q_OBJECT
+public:
+    enum Roles {
+        // Uses standard Qt::DisplayRole for text
+        ValueRole = Qt::UserRole + 1,
+    };
+
+    QHash<int, QByteArray> roleNames() const override;
+    int rowCount(const QModelIndex &parent) const override;
+    QVariant data(const QModelIndex &index, int role) const override;
+
+protected:
+    struct Entry {
+        int value;
+        QString display;
+    };
+
+    PolicyChoicesModel(QVector<Entry> &&policies, QObject *parent = nullptr);
+
+private:
+    const QVector<Entry> m_policies;
+};
+
+class FilesystemChoicesModel : public PolicyChoicesModel
+{
+    Q_OBJECT
+public:
+    FilesystemChoicesModel(QObject *parent = nullptr);
+};
+
+class DBusPolicyChoicesModel : public PolicyChoicesModel
+{
+    Q_OBJECT
+public:
+    DBusPolicyChoicesModel(QObject *parent = nullptr);
+};
+
 /**
  * @class FlatpakPermission describes a single configurable entry in the list model of permissions.
  *
@@ -56,6 +227,8 @@ public:
         Environment
     };
 
+    static ValueType valueTypeFromSectionType(FlatpakPermissionsSectionType::Type section);
+
     enum class OriginType {
         /**
          * Built-in type is for all pre-defined system resources (permissions)
@@ -87,32 +260,28 @@ public:
         Dummy
     };
 
-    enum class SectionType {
-        /**
-         * Easy-to-understand permissions, such as: print system access,
-         * internet connection, all filesystem resources etc.
-         */
-        Basic,
-        /**
-         * More "technical" permissions.
-         */
-        Advanced
-    };
+    using Variant = std::variant<QString, FlatpakPolicy>;
+
+    // Default constructor is required for meta-type registration.
+    /** Default constructor. Creates an invalid entry. */
+    FlatpakPermission() = default;
 
     /**
      * Create a Dummy entry for the Advanced and user-editable sections, just so
      * that ListView shows a section header even if there are no permission row
      * entries in it.
      */
-    explicit FlatpakPermission(const QString &category);
+    explicit FlatpakPermission(FlatpakPermissionsSectionType::Type section);
 
-    explicit FlatpakPermission(const QString &name,
+    explicit FlatpakPermission(FlatpakPermissionsSectionType::Type section,
+                               const QString &name,
                                const QString &category,
                                const QString &description,
-                               ValueType type,
                                bool isDefaultEnabled,
-                               const QString &defaultValue = {},
-                               const QStringList &possibleValues = {});
+                               const Variant &defaultValue = QString());
+
+    /** Section type for QtQuick/ListView. */
+    FlatpakPermissionsSectionType::Type section() const;
 
     /**
      * Technical untranslated name of the resource managed by this permission entry.
@@ -128,11 +297,6 @@ public:
      */
     const QString &category() const;
 
-    /**
-     * User-facing translated string used as section header.
-     */
-    QString categoryHeading() const;
-
     /**
      * Untranslate section heading back into category identifier. It's a hack
      * until the model is refactored to only operate on identifiers, and all
@@ -147,6 +311,9 @@ public:
      */
     const QString &description() const;
 
+    /**
+     * Return type of value of this entry, inferred from its SectionType.
+     */
     ValueType valueType() const;
 
     /**
@@ -160,18 +327,6 @@ public:
     // TODO: This method should be replaced with constructor argument.
     void setOriginType(OriginType type);
 
-    /**
-     * Section type of this permission. Defaults to SectionType::Basic for
-     * ValueType::Filesystems, and Advanced for everything else.
-     */
-    SectionType sectionType() const;
-
-    /**
-     * Set section type for this permission.
-     */
-    // TODO: This method should be replaced with constructor argument.
-    void setSectionType(SectionType type);
-
     /**
      * System default "enabled" status of this permission. It can not be modified.
      */
@@ -200,13 +355,13 @@ public:
      *
      * Applicable for any permissions other than ValueType::Simple.
      */
-    const QString &defaultValue() const;
+    const Variant defaultValue() const;
 
     // TODO: Remove this method.
-    const QString &overrideValue() const;
+    const Variant overrideValue() const;
 
     /** Set user override */
-    void setOverrideValue(const QString &value);
+    void setOverrideValue(const Variant &value);
 
     /**
      * This property holds the current effective value of this permission in
@@ -214,26 +369,21 @@ public:
      *
      * See ValueType enum for more.
      */
-    const QString &effectiveValue() const;
-    void setEffectiveValue(const QString &value);
+    const Variant effectiveValue() const;
+    void setEffectiveValue(const Variant &value);
 
     /** Untranslate value of ValueType::Filesystems permission. */
     // TODO: Remove this method, store enum variants or otherwise raw untranslated data.
     QString fsCurrentValue() const;
 
-    /**
-     * Provide a user-facing translated list of possible values for this kind of
-     * permission.
-     */
-    // TODO: It should be a model that also contains detailed description
-    // (help text) and untranslated value identifier.
-    const QStringList &possibleValues() const;
-
     /** Integration with KCM. */
     bool isSaveNeeded() const;
     bool isDefaults() const;
 
 private:
+    /** Section type for QtQuick/ListView. */
+    FlatpakPermissionsSectionType::Type m_section;
+
     /**
      * Untranslatable identifier of permission.
      *
@@ -250,9 +400,7 @@ private:
 
     /* Attempts to classify permissions into various types and groups. */
 
-    ValueType m_valueType;
     OriginType m_originType;
-    SectionType m_sectionType;
 
     /* Applicable for all ValueType permissions. */
 
@@ -266,16 +414,11 @@ private:
     /* Applicable for any permissions other than ValueType::Simple. */
 
     /** System defaults */
-    QString m_defaultValue;
+    Variant m_defaultValue;
     /** User overrides */
-    QString m_overrideValue;
+    Variant m_overrideValue;
     /** Current value in KCM */
-    QString m_effectiveValue;
-
-    /* Applicable for ValueType::Filesystems and ValueType::Bus only. */
-
-    /** Static list of translated policy names. */
-    QStringList m_possibleValues;
+    Variant m_effectiveValue;
 };
 
 class FlatpakPermissionModel : public QAbstractListModel
@@ -288,18 +431,20 @@ public:
     FlatpakPermissionModel(QObject *parent = nullptr);
 
     enum Roles {
-        Name = Qt::UserRole + 1, //
-        Category,
+        Section = Qt::UserRole + 1,
+        Name,
         Description,
         //
         IsSimple,
         IsEnvironment,
         IsNotDummy,
         //
-        IsGranted,
-        CurrentValue,
+        IsDefaultEnabled,
+        IsEffectiveEnabled,
+        DefaultValue,
+        EffectiveValue,
         //
-        ValueList,
+        ValuesModel,
     };
     Q_ENUM(Roles)
 
@@ -324,30 +469,47 @@ public:
     bool isDefaults() const;
     bool isSaveNeeded() const;
 
-    Q_INVOKABLE QStringList valueListForSection(const QString &sectionHeader) const;
-    Q_INVOKABLE QStringList valueListForUntranslatedCategory(const QString &category) const;
+    /**
+     * Static list of user-facing translated policy names for this kind of
+     * permission.
+     *
+     * Applicable for ValueType::Filesystems and ValueType::Bus only.
+     */
+    // TODO: It should be a model that also contains detailed description
+    // (help text) and untranslated value identifier.
+    Q_INVOKABLE static PolicyChoicesModel *valuesModelForSectionType(int /*FlatpakPermissionsSectionType::Type*/ rawSection);
+    Q_INVOKABLE static PolicyChoicesModel *valuesModelForFilesystemsSection();
+    Q_INVOKABLE static PolicyChoicesModel *valuesModelForBusSections();
+
+    Q_INVOKABLE static QString sectionHeaderForSectionType(int /*FlatpakPermissionsSectionType::Type*/ rawSection);
+    Q_INVOKABLE static QString sectionAddButtonToolTipTextForSectionType(int /*FlatpakPermissionsSectionType::Type*/ rawSection);
 
 public Q_SLOTS:
     void togglePermissionAtIndex(int index);
-    void editPerm(int index, const QString &newValue);
-    void addUserEnteredPermission(const QString &name, QString cat, const QString &value);
+    void editPerm(int index, const QVariant &newValue);
+    void addUserEnteredPermission(int /*FlatpakPermissionsSectionType::Type*/ rawSection, const QString &name, const QVariant &value);
 
 Q_SIGNALS:
     void referenceChanged();
     void showAdvancedChanged();
 
 private:
-    void addPermission(FlatpakPermission *perm, bool shouldBeOn);
-    void removePermission(FlatpakPermission *perm, bool isGranted);
-    void addBusPermissions(FlatpakPermission *perm);
-    void removeBusPermission(FlatpakPermission *perm);
-    void editFilesystemsPermissions(FlatpakPermission *perm, const QString &newValue);
-    void editBusPermissions(FlatpakPermission *perm, const QString &newValue);
-    void addEnvPermission(FlatpakPermission *perm);
-    void removeEnvPermission(FlatpakPermission *perm);
-    void editEnvPermission(FlatpakPermission *perm, const QString &newValue);
+    void addPermission(const FlatpakPermission &permission, bool shouldBeOn);
+    void removePermission(const FlatpakPermission &permission, bool isGranted);
+
+    void editFilesystemsPermissions(FlatpakPermission &permission, const QString &newValue);
+
+    void addBusPermissions(const FlatpakPermission &permission);
+    void removeBusPermission(const FlatpakPermission &permission);
+    void editBusPermissions(FlatpakPermission &permission, FlatpakPolicy newValue);
+
+    void addEnvPermission(const FlatpakPermission &permission);
+    void removeEnvPermission(const FlatpakPermission &permission);
+    void editEnvPermission(FlatpakPermission &permission, const QString &newValue);
+
     bool permExists(const QString &name) const;
     int permIndex(const QString &category);
+
     void readFromFile();
     void writeToFile();
 
